import re
from typing import Set

from prompt_toolkit.completion import (
    Completer,
    CompleteEvent,
    WordCompleter,
    NestedCompleter,
)
from prompt_toolkit.document import Document

from resotoshell.promptsession import (
    CommandLineCompleter,
    SearchCompleter,
    DocumentExtension,
    AggregateCompleter,
    PropertyListCompleter,
    CommandInfo,
    ArgInfo,
)

known_kinds = ["account", "instance", "instance_type", "network"]
known_props = [
    "age",
    "instance_memory",
    "instance_status",
    "some.nested.prop",
    "some.nested.other.prop",
    "some.nested.array[*].other[*].prop.foo",
]
known_commands = [
    CommandInfo("aggregate", source=False, args=[ArgInfo(None, True, value_hint="aggregate")]),
    CommandInfo(
        "ancestors",
        args=[
            ArgInfo("--with-origin", help_text="include the origin in the output"),
            ArgInfo(None, True, ["default", "delete"], help_text="edge type"),
        ],
        source=False,
    ),
    CommandInfo(
        "certificate",
        args={
            "create": [
                ArgInfo("--common-name", True, help_text="Common name like: example.com"),
                ArgInfo(
                    "--dns-names",
                    True,
                    help_text="List of other dns names: example.org example.io",
                ),
                ArgInfo(
                    "--ip-addresses",
                    True,
                    help_text="List of ip addresses: 1.2.3.4 2.3.4.5",
                ),
            ],
        },
    ),
    CommandInfo(
        "configs",
        args={
            "list": [],
            "set": [
                ArgInfo(None, expects_value=True, help_text="<config_id> <key>=<value>"),
            ],
            "show": [ArgInfo(None, expects_value=True, help_text="<config_id> e.g. resoto.core")],
            "edit": [
                ArgInfo(None, expects_value=True, help_text="<config_id> e.g. resoto.core"),
            ],
            "update": [
                ArgInfo(
                    None,
                    expects_value=True,
                    help_text="<config_id> /path/to/config.yaml",
                ),
            ],
            "delete": [
                ArgInfo(None, expects_value=True, help_text="<config_id> e.g. resoto.core"),
            ],
        },
    ),
    CommandInfo(
        "count",
        [
            ArgInfo(
                None,
                expects_value=True,
                help_text="optional property to count",
                value_hint="property",
            )
        ],
        source=False,
    ),
    CommandInfo(
        "list",
        [
            ArgInfo("--csv", help_text="format", option_group="format"),
            ArgInfo("--markdown", help_text="format", option_group="format"),
            ArgInfo(
                None,
                expects_value=True,
                help_text="the list of properties, comma separated",
                value_hint="property_list_with_as",
            ),
        ],
        source=False,
    ),
    CommandInfo(
        "search",
        args=[
            ArgInfo("--with-edges", help_text="include edges in the output"),
            ArgInfo(None, True, value_hint="search"),
        ],
    ),
]


def complete(part: str, completer: Completer, return_meta: bool = False) -> Set[str]:
    return {
        a.display_meta_text if return_meta else a.text
        for a in completer.get_completions(Document(part, len(part)), CompleteEvent())
    }


def test_doc_ext() -> None:
    d = Document("Sentence 1. Things I want to say", 32)
    ext = DocumentExtension(d, re.compile(r"\s*[.]\s*"))
    assert ext.last_doc() == Document("Things I want to say", 20)


def test_complete_command() -> None:
    n = CommandLineCompleter.create_completer(known_commands, [], [])

    # source commands
    assert len(complete("", n)) == 3
    assert complete("c", n) == {"certificate", "configs", "search"}
    assert complete("ce", n) == {"certificate"}

    # flow commands
    assert len(complete("search all | ", n)) == 4
    assert complete("search all | c", n) == {"ancestors", "count"}


def test_property() -> None:
    n = SearchCompleter(known_kinds, known_props)
    # /ancestors.xxx handling
    assert complete("/anc", n) == {"/ancestors"}
    assert len(complete("/ancestors.", n)) == len(known_kinds)
    assert complete("/ancestors.account.", n) == {"reported", "desired", "metadata"}
    assert len(complete("/ancestors.account.reported.", n)) >= len(known_props)
    assert complete("is(volume) and /anc", n) == {"/ancestors"}
    assert len(complete("is(volume) and /ancestors.", n)) == len(known_kinds)
    assert complete("is(volume) and /ancestors.account.", n) == {
        "reported",
        "desired",
        "metadata",
    }
    assert len(complete("is(volume) and /ancestors.account.reported.", n)) >= len(known_props)
    # /reported handling
    assert complete("/repo", n) == {"/reported"}
    assert len(complete("/reported.", n)) >= len(known_props)

    assert complete("some.nested.", n) == {
        "some.nested.array[*].other[*].prop.foo",
        "some.nested.other.prop",
        "some.nested.prop",
    }
    assert complete("some.nested.array[", n) == {"some.nested.array[*].other[*].prop.foo"}


def test_property_list() -> None:
    n = PropertyListCompleter(known_kinds, known_props)
    assert complete("/anc", n) == {"/ancestors"}
    assert complete("foo, /anc", n) == {"/ancestors"}
    assert complete("foo as bla, /anc", n) == {"/ancestors"}
    assert complete("foo as bla, ins", n) == {"instance_memory", "instance_status"}
    assert len(complete("foo as ", n)) == 1  # suggest name


def test_search() -> None:
    n = SearchCompleter(known_kinds, known_props)
    assert len(complete("", n)) > len(known_props)
    assert "is(" in complete("i", n)

    # show possible kinds in is()
    assert len(complete("is(", n)) == len(known_kinds)

    # show possible combinations after is()
    assert {"and", "or"} <= complete("is(foo) ", n)

    # show possible props for filter expression
    assert len(complete("is(instance) and ", n)) > len(known_props)
    assert len(complete("is(instance) or ", n)) > len(known_props)

    # show possible operators for filter expression
    assert {">", "<", "=", "!=", ">=", "<="} <= complete("is(instance) or age ", n)

    # show possible operators for filter expression
    assert complete("is(instance) or age >= ", n, True) == {"like 123, test, 12days, true, false, null, [1,2,3], {a:1}"}

    # suggest sort and limit after filter
    # show possible operators for filter expression
    assert {"sort", "limit"} <= complete("is(instance) or age >= 12d ", n)

    # show all properties for sorting
    assert len(complete("is(instance) or age >= 12d sort ", n)) >= len(known_props)

    # show sort order after the property
    assert {"asc", "desc"} <= complete("is(instance) or age >= 12d sort foo ", n)

    # do not suggest sort again, but limit
    assert "sort" not in complete("is(instance) or age >= 12d sort foo asc ", n)
    assert "limit" in complete("is(instance) or age >= 12d sort foo asc ", n)

    # suggest traverse operators after limit
    assert {"<--", "-->"} <= complete("is(instance) or age >= 12d sort foo limit 23, 12 ", n)

    # suggest filter expression after traverse
    assert {"is(", "all"} <= complete("is(instance) or age >= 12d sort foo limit 23, 12 --> ", n)
    assert {"is(", "all"} <= complete("(", n)


def test_aggregate() -> None:
    n = AggregateCompleter(known_kinds, known_props)
    # show all  possible properties (+ example etc)
    assert len(complete("", n)) >= len(known_props)
    # after the group
    assert {"as", ":", ","} <= complete("name ", n)
    # define a name for the group
    assert complete("name as ", n, True) == {"name of this result"}
    # after a complete group
    assert {":", ","} <= complete("name as bla ", n)
    assert "as" not in complete("name as bla ", n)
    # expect the next group var
    assert len(complete("name, ", n)) >= len(known_props)
    # works the same for multiple group vars
    assert {"as", ":", ","} <= complete("name, foo ", n)

    # show possible functions
    assert {"sum(", "min("} <= complete("name: ", n)
    # show all properties + static values
    assert len(complete("name: sum(", n)) >= len(known_props)
    # suggest to name this function
    assert {"as", ","} <= complete("name: sum(1)", n)
    # after named function, only comma is allowed
    assert complete("name: sum(1) as bla ", n) == {","}
    assert complete("name: sum(1) as bla, min(foo) as foo ", n) == {","}

    # start with function
    assert len(complete("sum(", n)) >= len(known_props)
    assert complete("sum(foo) as ", n, True) == {"name of this result"}
    assert {"sum(", "min("} <= complete("sum(foo) as foo,", n)


def test_complete_option() -> None:
    n = CommandLineCompleter.create_completer(known_commands, known_kinds, known_props)
    assert complete("ancestors ", n) == {
        "--with-origin",
        "default",
        "delete",
    }
    # show all options
    assert complete("ancestors --with-origin ", n) == {"default", "delete"}
    # once a direct value is supplied, no other option is shown
    assert complete("ancestors default ", n) == set()
    # once a direct value is supplied, no other option is shown
    assert complete("ancestors --with-origin default ", n) == set()
    assert complete("ancestors -", n) == {"--with-origin"}
    assert complete("configs ", n) == {
        "list",
        "set",
        "show",
        "edit",
        "update",
        "delete",
    }
    assert complete("configs show ", n, True) == {"<config_id> e.g. resoto.core"}
    assert complete("certificate create --common-name ", n, True) == {"Common name like: example.com"}
    assert "--common-name" not in complete("certificate create --common-name example.com ", n)
    assert complete("configs show ", n, True) == {"<config_id> e.g. resoto.core"}
    assert {"--markdown", "--csv"} <= complete("search all | list ", n)
    assert len(complete("search all | list ", n)) >= len(known_props)

    # nothing is suggested when the hint has been provided
    assert complete("echo hello ", n) == set()


def test_complete_word() -> None:
    w = WordCompleter(["aaa", "bbb", "ccc"])
    assert complete("", w) == {"aaa", "bbb", "ccc"}
    assert complete("a", w) == {"aaa"}

    n = NestedCompleter.from_nested_dict(
        {
            "aggregate": WordCompleter(["avg", "sum", "count"]),
            "ancestors": {
                "--with-origin": None,
                "default": {"foo": {"bla": {"bar"}}},
                "delete": None,
            },
        }
    )
    assert complete("", n) == {"aggregate", "ancestors"}

    # nested
    assert complete("anc", n) == {"ancestors"}
    assert complete("ancestors ", n) == {"--with-origin", "default", "delete"}
    assert complete("ancestors --", n) == {"--with-origin"}
    assert complete("ancestors default foo ", n) == {"bla"}
    assert complete("ancestors default foo bla ", n) == {"bar"}

    # words
    assert complete("aggregate ", n) == {"avg", "sum", "count"}
    assert complete("aggregate a", n) == {"avg"}



import os.path
import re
import shutil
import sys
from subprocess import call
from tempfile import TemporaryDirectory
from typing import Dict, Union, Optional, Tuple
from urllib.parse import urlsplit

from requests import Response
from requests.exceptions import ConnectionError
from requests_toolbelt import MultipartDecoder
from requests_toolbelt.multipart.decoder import BodyPart
from resotolib.args import ArgumentParser
from resotolib.utils import sha256sum
from resotolib.logger import log
from resotoshell.protected_files import validate_paths
from resotoclient import ResotoClient


class Shell:
    def __init__(
        self,
        client: ResotoClient,
        tty: bool,
        color_system: str,
        graph: Optional[str] = None,
        section: Optional[str] = None,
    ):
        self.client = client
        self.tty = tty
        self.color_system = color_system
        self.graph = graph
        self.section = section

    def handle_command(
        self,
        command: str,
        additional_headers: Optional[Dict[str, str]] = None,
        files: Optional[Dict[str, str]] = None,
    ) -> None:
        headers: Dict[str, str] = {}
        headers.update({"Accept": "text/plain"})
        headers.update(additional_headers or {})

        # set tty headers
        if self.tty:
            tty_columns, tty_rows = shutil.get_terminal_size(fallback=(80, 25))
            log.debug(f"Setting columns {tty_columns}, rows {tty_rows}")
            headers.update(
                {
                    "Resoto-Shell-Columns": str(tty_columns),
                    "Resoto-Shell-Rows": str(tty_rows),
                    "Resoto-Shell-Color-System": self.color_system,
                    "Resoto-Shell-Terminal": "true",
                }
            )

        def handle_response(maybe: Optional[Response], upload: bool = False) -> None:
            if maybe is not None:
                with maybe as response:
                    if response.status_code == 200:
                        self.handle_result(response)
                    elif response.status_code == 424 and not upload:
                        required = response.json().get("required", [])
                        to_upload = validate_paths({fp["name"]: fp["path"] for fp in required})
                        mp: Response = self.client.cli_execute_raw(
                            command=command,
                            files=to_upload,
                            graph=self.graph,
                            section=self.section,
                            headers=headers,
                        )
                        handle_response(mp, True)
                    else:
                        log.debug(f"HTTP error, code: {response.status_code}")
                        print(response.text, file=sys.stderr)
                        return

        try:
            received_response = self.client.cli_execute_raw(
                command=command,
                files=files,
                graph=self.graph,
                section=self.section,
                headers=headers,
            )
            handle_response(received_response)
        except ConnectionError:
            err = (
                "Error: Could not communicate with resotocore"
                f" at {urlsplit(self.client.resotocore_url).netloc}."
                " Is it up and reachable?"
            )
            print(err, file=sys.stderr)
        except Exception as ex:
            print(f"Error performing command: `{command}`\nReason: {ex}")

    def handle_result(self, part: Union[Response, BodyPart], first: bool = True) -> None:
        # store the file from the part inside the given directory
        def store_file(directory: str) -> Tuple[str, str]:
            disposition = part.headers.get("Content-Disposition", "")
            match = re.findall('filename="([^"]+)"', disposition)
            filename = match[0] if match else "out"
            if "/" in filename:
                raise ValueError(f"Invalid filename: {filename}")
            filepath = os.path.join(directory, filename)
            i = 0
            while os.path.exists(filepath):
                i += 1
                filepath = os.path.join(directory, f"{filename}-{i}")
            with open(filepath, "wb+") as fh:
                fh.write(part.content)
            return filename, filepath

        content_type = part.headers.get("Content-Type", "text/plain")
        action = part.headers.get("Resoto-Shell-Action")
        command = part.headers.get("Resoto-Shell-Command")
        line_delimiter = "---"

        # If we get a plain text result, we simply print it to the console.
        if content_type == "text/plain":
            # Received plain text: print it.
            if not first:
                print(line_delimiter)
            if hasattr(part, "iter_lines"):
                for line in part.iter_lines():
                    print(line.decode("utf-8"))
            else:
                print(part.text)
        # File is sent in order to edit and return it.
        # We expect the command to define what should happen with the edited file.
        elif content_type == "application/octet-stream" and action == "edit" and command:
            with TemporaryDirectory() as tmp:
                filename, filepath = store_file(tmp)
                original_shasum = sha256sum(filepath)
                call([os.environ.get("EDITOR", "vi"), filepath])
                new_shasum = sha256sum(filepath)
                log.debug(f"Original config sha256: {original_shasum}," f" new sha256: {new_shasum}")
                if new_shasum != original_shasum:
                    self.handle_command(f"{command} {filename}", {}, {filename: filepath})
                else:
                    print("No change made while editing the file. Update aborted.")
        # File is sent: save it to local disk
        elif content_type == "application/octet-stream":
            filename, filepath = store_file(ArgumentParser.args.download_directory)
            print(f"Received a file {filename}, which is stored to {filepath}.")
        # Multipart: handle each part separately
        elif content_type.startswith("multipart"):
            # Received a multipart response: parse the parts
            decoder = MultipartDecoder.from_response(part)

            def decode(value: Union[str, bytes]) -> str:
                return value.decode("utf-8") if isinstance(value, bytes) else value

            for num, part in enumerate(decoder.parts):
                part.headers = {decode(k): decode(v) for k, v in part.headers.items()}
                self.handle_result(part, num == 0)



import pathlib
import re
from abc import ABC
from attrs import define, field
from re import Pattern
from math import floor
from shutil import get_terminal_size
from typing import Iterable, Optional, List, Dict, Union, Tuple, Callable, Any

import jsons
from prompt_toolkit import PromptSession as PTSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import (
    Completer,
    CompleteEvent,
    Completion,
    WordCompleter,
    NestedCompleter,
    merge_completers,
    PathCompleter,
    FuzzyCompleter,
)
from prompt_toolkit.document import Document
from prompt_toolkit.history import FileHistory
from prompt_toolkit.styles import Style
from resotoclient import ResotoClient
from resotoclient.models import Property

from resotolib.logger import log


def cut_document_remaining(document: Document, span: Tuple[int, int]) -> Document:
    remaining = document.text_before_cursor[span[0] : span[1]]
    return Document(
        remaining,
        cursor_position=document.cursor_position - (len(document.text) - len(remaining)),
    )


def cut_document_last(document: Document, last_part: str) -> Document:
    text = document.text_before_cursor
    if last_part in text:
        left, right = document.text_before_cursor.rsplit(last_part, 1)
        right_stripped = right.lstrip()
        return Document(
            right_stripped,
            cursor_position=document.cursor_position - len(left) - len(last_part) - (len(right) - len(right_stripped)),
        )
    else:
        return document


@define
class ArgInfo:
    name: Optional[str]
    expects_value: bool = False
    possible_values: List[str] = field(factory=list)
    can_occur_multiple_times: bool = False
    value_hint: Optional[str] = None
    help_text: Optional[str] = None
    option_group: Optional[str] = None


# ArgsInfo = Union[Dict[str, Union["ArgsInfo", List[ArgInfo]]], List[ArgInfo]]
# mypy does not support recursive type aliases: define 3 levels as maximum here
ArgsInfo = Union[
    Dict[
        str,
        Union[
            Dict[str, Union[Dict[str, Union[Any, List[ArgInfo]]], List[ArgInfo]]],
            List[ArgInfo],
        ],
    ],
    List[ArgInfo],
]


@define
class CommandInfo:
    name: str
    args: ArgsInfo = field(factory=dict)  # type: ignore
    source: bool = True
    info: str = ""
    help: str = ""


@define
class ArgCompleter(Completer):
    arg: ArgInfo
    completer: Optional[Completer]

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        if self.completer:
            return self.completer.get_completions(document, complete_event)
        else:
            return []


@define
class CommandCompleter(Completer):
    cmd: CommandInfo
    completer: Optional[Completer]

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        if self.completer:
            return self.completer.get_completions(document, complete_event)
        else:
            return []


class FuzzyWordCompleter(Completer):
    def __init__(
        self,
        words: List[str],
        display_dict: Optional[Dict[str, str]] = None,
        meta_dict: Optional[Dict[str, str]] = None,
    ) -> None:
        self.words = words
        self.WORD = True
        self.word_completer = WordCompleter(
            words=self.words,
            WORD=self.WORD,
            display_dict=display_dict,
            meta_dict=meta_dict,
        )
        self.fuzzy_completer = FuzzyCompleter(self.word_completer, WORD=self.WORD)

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        return self.fuzzy_completer.get_completions(document, complete_event)

    def get_completions_filtered(
        self,
        document: Document,
        complete_event: CompleteEvent,
        predicate: Callable[[Completion], bool],
    ) -> Iterable[Completion]:
        for v in self.get_completions(document, complete_event):
            if predicate(v):
                yield v

    def get_completions_without(
        self, document: Document, complete_event: CompleteEvent, **filter_values: bool
    ) -> Iterable[Completion]:
        return self.get_completions_filtered(document, complete_event, lambda v: not filter_values.get(v.text, False))


re_nav_block = r"[a-zA-Z0-9]*(?:\[[0-9:]+\])?[a-zA-Z0-9]*"
re_navigation = re.compile(f"-{re_nav_block}->|<-{re_nav_block}-|<-{re_nav_block}->")

re_start_query = re.compile(r"^\s*(/?[A-Za-z0-9_.]*)$")
re_inside_is = re.compile(".*is\\(([^)]*)$")
re_json_value = r'(?:"[^"]*")|(?:\[[^\\]+\])|(?:\{[^}]+\})|(?:[A-Za-z0-9_\-:/.]+)'

re_partial_and_or = r"|a|an|and|o|or|-|<|-|-\[|--|<-|<-\[|s|so|sor|sort|l|li|lim|limi|limit"
re_param = "/?[A-Za-z_][A-Za-z0-9_\\-.\\[\\]]*"
re_op = "==|=|!=|<|>|<=|>=|=~|~|!~|\\s+in\\s+|\\s+not in\\s+"
re_fulltext = r'(?:"[^"]*")'

# match and/or
re_after_bracket = re.compile(f".*\\)\\s+({re_partial_and_or})$")
# noinspection RegExpUnnecessaryNonCapturingGroup
re_after_param_filter = re.compile(f".*{re_param}\\s*(?:{re_op})\\s*(?:{re_json_value})\\s({re_partial_and_or})$")
# noinspection RegExpUnnecessaryNonCapturingGroup
re_after_fulltext = re.compile(f".*(?:{re_fulltext})\\s*({re_partial_and_or})$")
re_after_sort_limit = re.compile(
    f".*(?:sort\\s+\\S+\\s(?:asc|desc)?|limit\\s+\\d+(?:,\\s*\\d+)?)\\s*({re_partial_and_or})$"
)
re_sort_attribute = re.compile(r".*sort\s+(\S*)$")
re_sort_order = re.compile(r".*sort\s+\S+\s+(\S*)$")

re_param_start = re.compile(r"(?:\s*|.*and\s+|.*or\s+)(/?[\w\-\[\]_.*]*)$")
re_slash_reported = re.compile(r"/reported.(\S*)$")
re_ancestor_descendant_kind = re.compile(r"(?:/ancestors.|/descendants.)([^.]*)$")
re_ancestor_descendant_section = re.compile(r"(?:/ancestors.|/descendants.)[\w\d_-]+[.]([^.]*)$")
re_ancestor_descendant_reported = re.compile(r"(?:/ancestors.|/descendants.)[\w\d_-]+[.]reported[.](\S*)$")
re_parm = "/?[A-Za-z0-9_\\-.\\[\\]*]*"
re_first_word = re.compile(f"^\\s*({re_parm})$")
re_second_word = re.compile(f"^\\s*{re_parm}\\s+(\\w*)$")
re_third_word = re.compile(f"^\\s*{re_parm}\\s+\\w+\\s+(\\w*)$")
re_after_third_word = re.compile(f"^\\s*{re_parm}\\s+\\w+\\s+\\w+\\s*(\\w*)$")
re_inside_function = re.compile(r".*\w+\(([^)]*)$")
re_fn = r"\w+\([^)]+\)"
re_second_word_after_fn = re.compile(f"^\\s*{re_fn}\\s*(\\w*)$")
re_third_word_after_fn = re.compile(f"^\\s*{re_fn}\\s+\\w+\\s*(\\w*)$")
re_after_third_word_fn = re.compile(f"^\\s*{re_fn}\\s+\\w+\\s+\\w+\\s*(\\w*)$")
re_after_bracket_start = re.compile(r"(?:|.*\s+)\((\w*)$")


class DocumentExtension:
    def __init__(self, document: Document, part_splitter: Optional[Pattern[str]] = None) -> None:
        self.document = document
        self.text = document.text_before_cursor
        self.parts = part_splitter.split(self.text) if part_splitter else [self.text]
        self.last = self.parts[-1].lstrip() if self.parts else ""
        self.last_words = self.last.split()
        self.last_word = self.word_at(-1)

    def cut_text(self, span: Tuple[int, int]) -> Document:
        return cut_document_remaining(self.document, span)

    def cut_last(self, span: Tuple[int, int]) -> Document:
        return cut_document_remaining(self.last_doc(), span)

    def last_doc(self) -> Document:
        return Document(self.last, self.document.cursor_position - (len(self.text) - len(self.last)))

    def word_at(self, idx: int) -> str:
        return self.last_words[idx] if abs(idx) <= len(self.last_words) else ""


class AbstractSearchCompleter(Completer, ABC):
    def __init__(self, kinds: List[str], props: List[str]):
        self.kinds = kinds
        self.props = props
        self.prop_lookup = set(props)
        ops = ["=", "!=", ">", "<", "<=", ">=", "~", "!~", "in"]
        self.ops_lookup = set(ops)
        self.kind_completer = FuzzyWordCompleter(kinds, meta_dict={p: "kind" for p in kinds})
        self.property_names_completer = FuzzyWordCompleter(
            props + ["/ancestors", "/reported", "/desired", "/metadata", "/descendants"],
            meta_dict=(
                {
                    "/reported": "absolute path in reported section",
                    "/desired": "absolute path in desired section",
                    "/metadata": "absolute path in metadata section",
                    "/ancestors": "ancestor properties",
                    "/descendants": "descendant properties",
                    **{p: "property" for p in self.props},
                }
            ),
        )
        self.start_completer = FuzzyWordCompleter(
            ['"', "is(", "/ancestors"] + props + ["/reported", "/desired", "/metadata", "/descendants", "all"],
            meta_dict=(
                {
                    '"': 'full text search. e.g. "test"',
                    "is(": "matches elements of defined kind",
                    "all": "matches all elements",
                    "/reported": "absolute path in reported section",
                    "/desired": "absolute path in desired section",
                    "/metadata": "absolute path in metadata section",
                    "/ancestors": "filter ancestor properties",
                    "/descendants": "filter descendant properties",
                    **{p: "filter property" for p in self.props},
                }
            ),
        )
        self.ops_completer = FuzzyWordCompleter(ops)
        self.and_or_completer = FuzzyWordCompleter(
            [
                "and",
                "or",
                "sort",
                "limit",
                "-->",
                "<--",
                "-[0:1]->",
                "<-[0:1]-",
                "-[0:]->",
                "<-[0:]-",
                "<-[0:]->",
            ],
            meta_dict={
                "and": "combine next term with and",
                "or": "combine next term with or",
                "sort": "sort results based on property",
                "limit": "limit results by offset and number",
                "-->": "traverse outbound one hop",
                "<--": "traverse inbound one hop",
                "-[0:1]->": "outbound one hop with current node",
                "<-[0:1]-": "inbound one hop with current node",
                "-[0:]->": "outbound to all leaves",
                "<-[0:]-": "inbound until the root",
                "<-[0:]->": "any reachable node from here",
            },
        )
        self.value_completer = HintCompleter("<value>", "like 123, test, 12days, true, false, null, [1,2,3], {a:1}")
        self.limit_completer = HintCompleter("<num>", "Number of elements with optional offset. e.g. 23 or 12, 23")
        self.sort_order_completer = FuzzyWordCompleter(["asc", "desc"])
        self.section_completer = FuzzyWordCompleter(["reported", "desired", "metadata"])

    def property_path_completions(
        self, completer: Completer, doc: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        all_completions = list(completer.get_completions(doc, complete_event))
        # reduce the list to the next dot, in case there are too many suggestions
        if len(all_completions) > 10:
            dl = len(doc.text_before_cursor)
            reduced = [a for a in all_completions if "." not in a.text[dl:]]
            all_completions = reduced if reduced else all_completions
        return all_completions

    def property_completions(
        self,
        doc: Document,
        ext: DocumentExtension,
        complete_event: CompleteEvent,
        start_with: Completer,
    ) -> Iterable[Completion]:
        if kind := re_ancestor_descendant_kind.match(ext.last_word):
            pd = cut_document_remaining(doc, kind.span(1))
            return self.kind_completer.get_completions(pd, complete_event)
        elif section := re_ancestor_descendant_section.match(ext.last_word):
            pd = cut_document_remaining(doc, section.span(1))
            return self.section_completer.get_completions(pd, complete_event)
        elif section := re_ancestor_descendant_reported.match(ext.last_word):
            pd = cut_document_remaining(doc, section.span(1))
            return self.property_path_completions(self.property_names_completer, pd, complete_event)
        elif reported := re_slash_reported.match(ext.last_word):
            pd = cut_document_remaining(doc, reported.span(1))
            return self.property_path_completions(self.property_names_completer, pd, complete_event)
        else:
            return self.property_path_completions(start_with, doc, complete_event)


class KindCompleter(AbstractSearchCompleter):
    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        return self.kind_completer.get_completions(document, complete_event)


class SearchCompleter(AbstractSearchCompleter):
    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        ext = DocumentExtension(document, re_navigation)
        have_sort = " sort " in ext.last
        have_limit = " limit " in ext.last

        if in_start := re_start_query.match(ext.last):
            doc = ext.cut_last(in_start.span(1))
            return self.property_completions(doc, ext, complete_event, self.start_completer)
        elif parm := re_param_start.match(ext.text):
            doc = ext.cut_text(parm.span(1))
            return self.property_completions(doc, ext, complete_event, self.start_completer)
        elif bracket := re_after_bracket_start.match(ext.text):
            doc = ext.cut_text(bracket.span(1))
            return self.property_completions(doc, ext, complete_event, self.start_completer)
        elif in_is := re_inside_is.match(ext.text):
            doc = ext.cut_text(in_is.span(1))
            return self.kind_completer.get_completions(doc, complete_event)
        elif after := re_after_bracket.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.and_or_completer.get_completions(doc, complete_event)
        elif after := re_after_fulltext.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.and_or_completer.get_completions(doc, complete_event)
        elif after := re_sort_attribute.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.property_completions(doc, ext, complete_event, self.property_names_completer)
        elif after := re_sort_order.match(ext.last):  # sort order
            doc = ext.cut_last(after.span(1))
            return self.sort_order_completer.get_completions(doc, complete_event)
        elif ext.last_word == "limit":
            doc = cut_document_last(document, ext.last_word)
            return self.limit_completer.get_completions(doc, complete_event)
        elif after := re_after_sort_limit.match(ext.last):
            doc = cut_document_remaining(document, after.span(1))
            return self.and_or_completer.get_completions_without(
                doc,
                complete_event,
                **{
                    "sort": have_sort | have_limit,
                    "limit": have_limit,
                    "and": have_sort | have_limit,
                    "or": have_sort | have_limit,
                },
            )
        elif after := re_after_param_filter.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.and_or_completer.get_completions(doc, complete_event)
        elif (ext.last_word in self.prop_lookup or ext.last_word.startswith("/")) and not have_sort:
            doc = cut_document_last(document, ext.last_word)
            return self.ops_completer.get_completions(doc, complete_event)
        elif ext.last_word in self.ops_lookup:
            doc = cut_document_last(document, ext.last_word)
            return self.value_completer.get_completions(doc, complete_event)
        else:
            return []


class AggregateCompleter(AbstractSearchCompleter):
    def __init__(self, kinds: List[str], props: List[str]) -> None:
        super().__init__(kinds, props)
        self.aggregate_fns = ["sum(", "min(", "max(", "avg("]
        self.aggregate_fn_completer = FuzzyWordCompleter(
            self.aggregate_fns,
            meta_dict={
                "sum(": "sum over all occurrences",
                "min(": "use the smallest occurrence",
                "max(": "use the biggest occurrence",
                "avg(": "average over all occurrences",
            },
        )
        self.as_completer = FuzzyWordCompleter(["as"], meta_dict=({"as": "rename this result"}))
        self.colon_completer = FuzzyWordCompleter(
            [":"],
            meta_dict=({":": "to define functions for this group"}),
        )
        self.comma_var_completer = FuzzyWordCompleter(
            [","],
            meta_dict=({",": "define another group variable"}),
        )
        self.comma_fn_completer = FuzzyWordCompleter(
            [","],
            meta_dict=({",": "define another group function"}),
        )
        self.props_completer = FuzzyWordCompleter(
            props + ["/ancestors", "/reported", "/desired", "/metadata", "/descendants"],
            meta_dict=(
                {
                    "/reported": "absolute path in reported section",
                    "/desired": "absolute path in desired section",
                    "/metadata": "absolute path in metadata section",
                    "/ancestors": "on ancestor properties",
                    "/descendants": "on descendant properties",
                    **{p: "aggregate property" for p in self.props},
                }
            ),
        )
        self.group_with_value_completer = merge_completers(
            [HintCompleter("1", "Static value to count", "1"), self.props_completer]
        )
        self.group_after_name = merge_completers([self.as_completer, self.colon_completer, self.comma_var_completer])
        self.fn_after_name = merge_completers([self.as_completer, self.comma_fn_completer])
        self.after_group = merge_completers([self.comma_var_completer, self.colon_completer])
        self.after_fn_completer = merge_completers([self.as_completer, self.comma_fn_completer])
        self.hint_completer = HintCompleter(
            "Example",
            "aggregation example",
            "kind, volume_type as type: sum(volume_size) as volume_size",
        )
        self.with_hint_completer = merge_completers(
            [self.hint_completer, self.props_completer, self.aggregate_fn_completer]
        )
        self.value_hint_completer = HintCompleter("<name>", "name of this result")

    def group_completion(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        ext = DocumentExtension(document, re.compile(r"\s*,\s*"))
        if first := re_first_word.match(ext.last):
            doc = ext.cut_last(first.span(1))
            return self.property_completions(doc, ext, complete_event, self.props_completer)
        elif second := re_second_word.match(ext.last):
            doc = ext.cut_last(second.span(1))
            return self.group_after_name.get_completions(doc, complete_event)
        elif third := re_third_word.match(ext.last):
            doc = ext.cut_last(third.span(1))
            return self.value_hint_completer.get_completions(doc, complete_event)
        elif after := re_after_third_word.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.after_group.get_completions(doc, complete_event)
        else:
            return []

    def fn_completion(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        ext = DocumentExtension(cut_document_last(document, ":"), re.compile(r"\s*,\s*"))
        if first := re_first_word.match(ext.last):
            doc = ext.cut_last(first.span(1))
            return self.aggregate_fn_completer.get_completions(doc, complete_event)
        elif in_fn := re_inside_function.match(ext.last):
            doc = ext.cut_last(in_fn.span(1))
            return self.group_with_value_completer.get_completions(doc, complete_event)
        elif second := re_second_word_after_fn.match(ext.last):
            doc = ext.cut_last(second.span(1))
            return self.fn_after_name.get_completions(doc, complete_event)
        elif third := re_third_word_after_fn.match(ext.last):
            doc = ext.cut_last(third.span(1))
            return self.value_hint_completer.get_completions(doc, complete_event)
        elif after := re_after_third_word_fn.match(ext.last):
            doc = ext.cut_last(after.span(1))
            return self.comma_fn_completer.get_completions(doc, complete_event)
        else:
            return []

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        text = document.text_before_cursor
        in_functions = ":" in text or any(a for a in self.aggregate_fns if a in text)
        if text.strip() == "":
            return self.with_hint_completer.get_completions(document, complete_event)
        elif in_functions:
            return self.fn_completion(document, complete_event)
        else:
            return self.group_completion(document, complete_event)


class PropertyCompleter(AbstractSearchCompleter):
    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        ext = DocumentExtension(document)
        return self.property_completions(document, ext, complete_event, self.property_names_completer)


class PropertyListCompleter(AggregateCompleter):
    # Inherit from aggregate completer only, since the grouping completer
    # already implements everything we need.
    def __init__(self, kinds: List[str], props: List[str], with_as: bool = True) -> None:
        super().__init__(kinds, props)
        self.comma_var_completer = FuzzyWordCompleter(
            [","],
            meta_dict=({",": "define another variable"}),
        )
        self.props_completer = self.property_names_completer
        self.group_after_name = (
            merge_completers([self.as_completer, self.comma_var_completer]) if with_as else self.comma_var_completer
        )
        self.after_group = self.comma_var_completer

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        return self.group_completion(document, complete_event)


class HintCompleter(Completer):
    def __init__(self, hint: str, meta: Optional[str] = None, value: str = " ") -> None:
        self.value = value
        self.cpl = WordCompleter([value], meta_dict={value: meta}, display_dict={value: hint}, WORD=True)

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        # Only show a hint once at the start
        if self.value == " " and document.text_before_cursor.strip() != "":
            return []
        else:
            return self.cpl.get_completions(document, complete_event)


class ArgsCompleter(Completer):
    def __init__(self, args: List[ArgCompleter]):
        direct: List[ArgCompleter] = []
        opts: List[ArgCompleter] = []
        for x in args:
            (direct if x.arg.name is None else opts).append(x)
        self.direct = direct
        self.args: Dict[str, ArgCompleter] = {arg.arg.name: arg for arg in opts if arg.arg.name is not None}
        self.value_lookup: Dict[str, ArgCompleter] = {v: a for a in args for v in a.arg.possible_values}
        self.options_completer = FuzzyWordCompleter(
            list(self.args.keys()),
            meta_dict={
                arg: complete.arg.help_text for arg, complete in self.args.items() if complete.arg.help_text is not None
            },
        )

    def inside_option(self, parts: List[str]) -> Optional[ArgCompleter]:
        for idx, part in enumerate(reversed(parts)):
            if (parm := self.args.get(part)) is not None:
                return parm if parm.arg.expects_value and idx < 2 else None
        return None

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        text = document.text_before_cursor
        parts = text.strip().split()
        maybe = self.inside_option(parts)

        # eat all options: remaining parts can be completed eventually
        adapted = text
        for arg in self.args.values():
            val_pattern = "\\s+\\S+\\s+" if arg.arg.expects_value else "\\s*"
            adapted = re.sub(f"\\s*{arg.arg.name}{val_pattern}", "", adapted)
        doc = Document(
            adapted,
            cursor_position=document.cursor_position - (len(text) - len(adapted)),
        )

        # look at the last word to see if an option is started
        adapted_stripped = adapted.lstrip()
        sp = adapted_stripped.rsplit(maxsplit=1)
        last = sp[-1] if len(sp) > 1 else adapted_stripped
        start_arg = last.strip() != "" and any(a.arg.name.startswith(last) for a in self.args.values() if a.arg.name)

        def direct_completers() -> Completer:
            def allowed_arg(arg_info: ArgInfo) -> bool:
                group = arg_info.option_group
                return group is None or not any(
                    a.arg.name in parts for a in self.args.values() if a.arg.option_group == group
                )

            d = [a.completer for a in self.direct if a.completer if allowed_arg(a.arg)]
            return merge_completers(d)

        def allowed(completion: Completion) -> bool:
            txt = completion.text
            # has this parameter been specified before?
            already_defined = txt in parts and txt in self.args and not self.args[txt].arg.can_occur_multiple_times

            # is this a possible value, where another value is already defined?
            another_value_defined = False
            if txt in self.value_lookup:
                another_value_defined = any(v for v in self.value_lookup[txt].arg.possible_values if v in parts)

            # another option group defined
            ag = self.args.get(completion.text)
            option_group_defined = False
            if group := ag.arg.option_group if ag else None:
                option_group_defined = any(
                    a.arg.name in parts for a in self.args.values() if a.arg.option_group == group
                )

            # if we come here: this is a valid completion
            return False if already_defined or another_value_defined or option_group_defined else True

        # either there is no option or an option has been started
        if adapted_stripped == "" or start_arg:
            last_doc = Document(
                last,
                cursor_position=document.cursor_position - (len(document.text) - len(last)),
            )
            direct = direct_completers().get_completions(doc, complete_event)
            opts = self.options_completer.get_completions(last_doc, complete_event)
            return [c for c in opts if allowed(c)] + list(direct)
            # inside an option
        elif maybe is not None and maybe.arg.name is not None:
            # suggest the arg values
            doc = cut_document_last(doc, maybe.arg.name)
            return maybe.get_completions(doc, complete_event)
            # no option started and not inside any option: assume a direct completion
        else:
            return direct_completers().get_completions(doc, complete_event)


class CommandLineCompleter(Completer):
    def __init__(self, commands: List[CommandInfo], completers: List[CommandCompleter]):
        self.commands = commands
        self.command_completers = {c.cmd.name: c for c in completers}
        self.source_completer = FuzzyWordCompleter([c.name for c in commands if c.source])
        self.flow_completer = FuzzyWordCompleter([c.name for c in commands if not c.source])

    def find_current_command(self, document: Document) -> Optional[Tuple[CommandCompleter, str]]:
        parts = document.text_before_cursor.lstrip().split(maxsplit=1)
        if parts and (command := self.command_completers.get(parts[0])) is not None:
            args = parts[1] if len(parts) == 2 else ""
            return command, args
        return None

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        text = document.text_before_cursor.lstrip()

        adapted = document
        if ";" in text:
            left, right = document.text_before_cursor.rsplit(";", 1)
            adapted = Document(right, document.cursor_position - len(left) - 1)

        if "|" in adapted.text_before_cursor:
            left, right = adapted.text_before_cursor.rsplit("|", 1)
            adapted = Document(right, document.cursor_position - len(left) - 1)

        if not text.endswith((";", "|")):  # enforce a space after ; or |
            maybe = self.find_current_command(adapted)
            if maybe is not None:
                completer, arg = maybe
                doc = Document(arg, len(arg))
                return completer.get_completions(doc, complete_event)
            elif ";" in text or "|" in text:
                return self.flow_completer.get_completions(adapted, complete_event)
            else:
                return self.source_completer.get_completions(adapted, complete_event)
        else:
            return []

    @staticmethod
    def create_completer(cmds: List[CommandInfo], kinds: List[str], props: List[str]) -> "CommandLineCompleter":
        def arg_completer(arg: ArgInfo) -> Optional[Completer]:
            if arg.possible_values:
                meta = {a: arg.help_text for a in arg.possible_values} if arg.help_text is not None else None
                return FuzzyWordCompleter(arg.possible_values, meta_dict=meta)
            elif arg.value_hint == "file":
                return PathCompleter()
            elif arg.value_hint == "kind":
                return KindCompleter(kinds, props)
            elif arg.value_hint == "property":
                return PropertyCompleter(kinds, props)
            elif arg.value_hint == "property_list_plain":
                return PropertyListCompleter(kinds, props, with_as=False)
            elif arg.value_hint == "property_list_with_as":
                return PropertyListCompleter(kinds, props, with_as=True)
            elif arg.value_hint == "command":
                meta = {cmd.name: "command" for cmd in cmds}
                return FuzzyWordCompleter([cmd.name for cmd in cmds], meta_dict=meta)
            elif arg.value_hint == "search":
                return SearchCompleter(kinds, props)
            elif arg.value_hint == "aggregate":
                return AggregateCompleter(kinds, props)
            elif arg.help_text:
                return HintCompleter("<value>", arg.help_text)
            else:
                return HintCompleter("<value>")

        def arg_value_completer(arg: ArgInfo) -> ArgCompleter:
            return ArgCompleter(arg, arg_completer(arg))

        def command_completer(cmd: CommandInfo) -> Optional[Completer]:
            def walk_commands(sub_commands: Dict[str, Union["ArgsInfo", List[ArgInfo]]]) -> Completer:
                return NestedCompleter.from_nested_dict(
                    {
                        sub_cmd: ArgsCompleter([arg_value_completer(arg) for arg in options])
                        if isinstance(options, list)
                        else walk_commands(options)
                        for sub_cmd, options in sub_commands.items()
                    }
                )

            if isinstance(cmd.args, list):
                return ArgsCompleter([arg_value_completer(arg) for arg in cmd.args])
            elif isinstance(cmd.args, dict):
                return walk_commands(cmd.args)
            else:
                return None

        cmd_completer = [CommandCompleter(cmd, command_completer(cmd)) for cmd in cmds]
        return CommandLineCompleter(cmds, cmd_completer)


class SafeCompleter(Completer):
    """
    This completer ensures, that the underlying completer will always return "something"
    prompt_toolkit does not handle empty completions very well.
    """

    def __init__(self, completer: Completer) -> None:
        self.completer = completer

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        try:
            result = self.completer.get_completions(document, complete_event)
            return [] if result is None else result
        except Exception as ex:
            log.warning(f"Error in completer: {ex}", exc_info=ex)
            return []


class PromptSession:

    style = Style.from_dict(
        {
            "green": "#00ff00",
            "red": "#ff0000",
            "prompt": "#C724B1",
        }
    )

    prompt_message = [("class:prompt", "> ")]

    def __init__(
        self,
        client: ResotoClient,
        history_file: str = str(pathlib.Path.home() / ".resotoshell_history"),
    ):
        history = FileHistory(history_file)
        cmds, kinds, props = self.__core_metadata(client)
        _, tty_rows = get_terminal_size(fallback=(80, 25))
        reserved_row_ratio = 1 / 4
        min_reserved_rows = 4
        max_reserved_rows = 12
        reserved_space_for_menu = max(min_reserved_rows, min(floor(tty_rows * reserved_row_ratio), max_reserved_rows))
        self.completer = CommandLineCompleter.create_completer(cmds, kinds, props)
        self.session: PTSession[str] = PTSession(history=history, reserve_space_for_menu=reserved_space_for_menu)

    def prompt(self) -> str:
        return self.session.prompt(
            self.prompt_message,  # type: ignore
            completer=SafeCompleter(self.completer),
            complete_while_typing=True,
            style=self.style,
            auto_suggest=AutoSuggestFromHistory(),
        )

    @staticmethod
    def __core_metadata(
        client: ResotoClient,
    ) -> Tuple[List[CommandInfo], List[str], List[str]]:

        try:
            log.debug("Fetching core metadata..")
            model = client.model()

            def path(p: Property) -> List[str]:
                kind = p.kind
                name = p.name
                result = [name]
                if p.kind.endswith("[]"):
                    kind = kind[:-2]
                    name += "[*]"
                    result.append(name)

                kd = model.kinds.get(kind)
                if kd is not None and kd.properties:
                    result.extend(name + "." + pp for prop in kd.properties for pp in path(prop))
                return result

            aggregate_roots = {
                k: v for k, v in model.kinds.items() if getattr(v, "aggregate_root", True) and v.properties is not None
            }

            known_props = {p for v in aggregate_roots.values() for prop in v.properties or [] for p in path(prop)}
            info = client.cli_info()
            cmds = [jsons.load(cmd, CommandInfo) for cmd in info.get("commands", [])]
            return cmds, sorted(aggregate_roots.keys()), sorted(known_props)
        except Exception as ex:
            log.warning(
                f"Can not load metadata from core: {ex}. No suggestions as fallback.",
                exc_info=ex,
            )
            return [], [], []



import logging
import os.path
import platform
import re
from re import RegexFlag
from typing import Dict

log = logging.getLogger(__name__)

# OS -> List[regexp_strings]
forbidden = {
    "windows": [".*\\/windows\\/.*", ".*SystemRoot.*"],
    "darwin": ["\\/System.*", "\\/Library.*"],
    "linux": [".*\\/etc\\/.*"],
}


def validate_paths(required: Dict[str, str]) -> Dict[str, str]:
    result = {}
    for name, path_in in required.items():
        path = os.path.abspath(os.path.expanduser(path_in))
        if not os.path.exists(path):
            raise AttributeError(f"Path does not exist: {path}")
        if not os.path.isfile(path):
            raise AttributeError(f"Path is not a file: {path}")
        if is_protected_file(path):
            raise AttributeError(f"Not allowed to upload {path}. This path is protected.")
        result[name] = path
    return result


def is_protected_file(path: str, system: str = platform.system()) -> bool:
    """
    Check if given path points to a file that should not be visible to anybody else.
    :param path: the path to check.
    :param system: only here for testing purposes.
    :return: True if the file is protected, otherwise false.
    """
    path_lower = os.path.abspath(path).lower()
    for check in forbidden.get(system.lower(), []):
        if re.compile(check, RegexFlag.IGNORECASE).fullmatch(path_lower):
            return True
    return False



import os
import sys
from argparse import Namespace
from signal import SIGTERM
from threading import Event, Thread
from typing import Callable

import resotolib.proc
from resotoclient import ResotoClient
from resotolib.args import ArgumentParser
from resotolib.core import resotocore, add_args as core_add_args, resotocore_is_up
from resotolib.core.ca import TLSData
from resotolib.event import add_event_listener, Event as ResotoEvent, EventType
from resotolib.jwt import add_args as jwt_add_args
from resotolib.logger import log, setup_logger, add_args as logging_add_args
from resotoshell.promptsession import PromptSession
from resotoshell.shell import Shell
from rich.console import Console


def main() -> None:
    resotolib.proc.parent_pid = os.getpid()
    resotolib.proc.initializer()
    setup_logger("resotoshell", json_format=False)
    arg_parser = ArgumentParser(description="resoto shell", env_args_prefix="RESOTOSHELL_")
    core_add_args(arg_parser)
    add_args(arg_parser)
    logging_add_args(arg_parser)
    jwt_add_args(arg_parser)
    TLSData.add_args(arg_parser, ca_only=True)
    args: Namespace = arg_parser.parse_args()

    if not resotocore_is_up(resotocore.http_uri):
        log.fatal(f"resotocore is not online at {resotocore.http_uri}")
        sys.exit(1)

    client = ResotoClient(
        url=resotocore.http_uri,
        psk=args.psk,
        custom_ca_cert_path=args.ca_cert,
        verify=args.verify_certs,
    )

    def check_system_info() -> None:
        try:
            list(client.cli_execute("system info"))
        except Exception as e:
            log.error(f"resotocore is not accessible: {e}")
            raise e

    try:
        client.start()
        check_system_info()
    except Exception:
        client.shutdown()
        sys.exit(1)
    if args.stdin or not sys.stdin.isatty():
        handle_from_stdin(client)
    else:
        repl(client, args)
    client.shutdown()
    resotolib.proc.kill_children(SIGTERM, ensure_death=True)
    log.debug("Shutdown complete")
    sys.exit(0)


def repl(
    client: ResotoClient,
    args: Namespace,
) -> None:
    shutdown_event = Event()
    shell = Shell(client, True, detect_color_system(args))
    session = PromptSession(client)
    log.debug("Starting interactive session")

    def shutdown(event: ResotoEvent) -> None:
        shutdown_event.set()
        kt = Thread(target=resotolib.proc.delayed_exit, name="shutdown")
        kt.start()

    add_event_listener(EventType.SHUTDOWN, shutdown)

    # send the welcome command to the core
    shell.handle_command("welcome")
    while not shutdown_event.is_set():
        try:
            command = session.prompt()
            if command == "":
                continue
            if command == "quit":
                shutdown_event.set()
                continue
            shell.handle_command(command)
        except KeyboardInterrupt:
            pass
        except EOFError:
            shutdown_event.set()
        except (RuntimeError, ValueError) as e:
            log.error(e)
        except Exception:
            log.exception("Caught unhandled exception while processing CLI command")


def handle_from_stdin(client: ResotoClient) -> None:
    shell = Shell(client, False, "monochrome")
    log.debug("Reading commands from STDIN")
    try:
        for command in sys.stdin.readlines():
            command = command.rstrip()
            shell.handle_command(command)
    except KeyboardInterrupt:
        pass
    except (RuntimeError, ValueError) as e:
        log.error(e)
    except Exception:
        log.exception("Caught unhandled exception while processing CLI command")


def detect_color_system(args: Namespace) -> str:
    if args.no_color:
        return "monochrome"
    else:
        lookup = {
            None: "monochrome",
            "standard": "standard",
            "256": "eight_bit",
            "truecolor": "truecolor",
            "windows": "legacy_windows",
        }
        cs = lookup.get(Console().color_system, "standard")
        log.debug(f"Detected color system is: {cs}")
        return cs


def add_args(arg_parser: ArgumentParser) -> None:
    def is_file(message: str) -> Callable[[str], str]:
        def check_file(path: str) -> str:
            if os.path.isfile(path):
                return path
            else:
                raise AttributeError(f"{message}: path {path} is not a directory!")

        return check_file

    arg_parser.add_argument(
        "--resotocore-section",
        help="All queries are interpreted with this section name. If not set, the server default is used.",
        dest="resotocore_section",
    )
    arg_parser.add_argument(
        "--resotocore-graph",
        help="The name of the graph to use by default. If not set, the server default is used.",
        dest="resotocore_graph",
    )
    arg_parser.add_argument(
        "--download-directory",
        help="If files are received, they are written to this directory.",
        default=".",
        dest="download_directory",
    )
    arg_parser.add_argument(
        "--no-color",
        help="Output should be rendered plain without any color escape sequences.",
        dest="no_color",
        action="store_true",
        default=False,
    )
    arg_parser.add_argument(
        "--stdin",
        help="Read from STDIN instead of opening a shell",
        dest="stdin",
        action="store_true",
        default=False,
    )


if __name__ == "__main__":
    main()



import os

import pytest

from resotoshell.protected_files import is_protected_file, validate_paths


def test_validate_path() -> None:
    # to test for files, pick one existing in cwd
    cwd_files = [a for a in os.listdir(".") if os.path.isfile(a)]
    if cwd_files:
        first = cwd_files[0]
        validated = validate_paths({"file": f"./{os.path.basename(first)}"})
        assert validated["file"] == os.path.abspath(first)

    with pytest.raises(AttributeError):  # unknown path
        validate_paths({"file": "/this/path/does/not/exist"})
    with pytest.raises(AttributeError):  # not a file
        validate_paths({"file": "/"})


def test_is_protected_file() -> None:
    assert is_protected_file("C:/windows/system32/some/oath", "windows")
    assert is_protected_file("%SystemRoot%/some/path", "windows")
    assert not is_protected_file("D:/application/test", "windows")

    assert is_protected_file("/System/Test/File", "darwin")
    assert is_protected_file("/Library/Test/File", "darwin")
    assert not is_protected_file("/Users/foo/file", "darwin")

    assert is_protected_file("/etc/passwd", "linux")
    assert not is_protected_file("/opt/test/file", "linux")
    assert not is_protected_file("/home/foo/file", "darwin")
